from pyskema import __version__

project = 'Pyskema'
copyright = '2023, Théo Cavignac'
author = 'Théo Cavignac'
release = __version__

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



"""Schema base visitor.

:mod:`pyskema` uses the `visitor patter <https://en.wikipedia.org/wiki/Visitor_pattern>`_
to implement various operations on schema.
This module implement the base class that should be extended to implement such operation.
"""
from .schema import Node


class Visitor:
    "The base visitor."

    def visit(self, node: Node, *args, **kwargs):
        "The entry point for all :class:`scheme.Node`."
        return node.structure.accept(self, *args, **kwargs)

    def visit_atom(self, atom, *args, **kwargs):
        "Visit an :class:`pyskema.schema.Atom` instance."
        pass

    def visit_union(self, union, *args, **kwargs):
        "Visit an :class:`pyskema.schema.Union` instance."
        for node in union.options:
            self.visit(node, *args, **kwargs)

    def visit_record(self, rec, *args, **kwargs):
        "Visit a :class:`pyskema.schema.Record` instance."
        for _, node in rec.fields.items():
            self.visit(node, *args, **kwargs)

    def visit_collection(self, collection, *args, **kwargs):
        "Visit a :class:`pyskema.schema.Collection` instance."
        self.visit(collection.element)

    def visit_map(self, map_, *args, **kwargs):
        "Visit a :class:`pyskema.schema.Map` instance."
        self.visit(map_.element)

    def visit_tuple(self, tup, *args, **kwargs):
        "Visit a :class:`pyskema.schema.Tuple` instance."
        for node in tup.fields:
            self.visit(node, *args, **kwargs)



"Implement validation of data against a schema."

from dataclasses import dataclass
from typing import List


from .schema import AtomType, Atom, Node, Record, Collection
from .visitor import Visitor
from .misc import closest, Result


def validate(data, schema, fail=True):
    """Validate a piece of data against a schema.

    :param data: a piece of data to validate.
    :param schema: a schema to refer to
    :param fail: (optional, :code:`True`)

        - if :code:`True`, raise an :class:`InvalidDataError` when failing to validate data
        - else, return None
    """
    assert isinstance(schema, Node), f"schema should be a Node instance, not {schema}."

    r = Validator.validate(schema, data)

    if r.valid:
        return r.value
    elif fail:
        raise InvalidDataError(r.error)

    return None


class Validator(Visitor):
    """Visitor implementing validation.

    You should probably be using :func:`validate` instead.
    """

    @classmethod
    def validate(cls, schema, data):
        """validate data against schema.

        :param schema: instance of Node
        :param data: the data to validate
        """
        return cls().visit(schema, data)

    def visit_atom(self, atom, data) -> Result:
        if atom.type_ == AtomType.INT:
            if isinstance(data, int):
                return Result.ok(data)
            elif isinstance(data, float):
                if int(data) == data:
                    return Result.ok(int(data))

        elif atom.type_ == AtomType.FLOAT:
            if isinstance(data, (float, int)):
                return Result.ok(data)

        elif atom.type_ == AtomType.STR:
            if isinstance(data, str):
                return Result.ok(data)

        elif atom.type_ == AtomType.OPTION:
            assert atom.options is not None, "atom.options must be set"
            if isinstance(data, str) and data in atom.options:
                return Result.ok(data)

        elif atom.type_ == AtomType.BOOL:
            if isinstance(data, bool):
                return Result.ok(data)

        else:
            raise NotImplementedError(f"{atom.type_} is not supported yet.")

        # invalid atom
        return Result.fail(
            Mismatch(
                [],
                data,
                f"{repr(data)} is not a valid {atom.type_.name(atom.options)}.",
            )
        )

    def visit_union(self, union, data) -> Result:
        errors = []
        for op in union.options:
            m = self.visit(op, data)
            if m.valid:
                # success at first match
                return Result.ok((op, m.value))
            else:
                errors.append(m.error)

        # no valid option
        return Result.fail(
            Mismatch(
                [],
                data,
                f"{repr(data)} does not validate any of the valid alternatives:\n"
                + "\n".join("- " + error_message(err, no_loc=True) for err in errors),
            )
        )

    def visit_record(self, rec, data) -> Result:
        try:
            pairs = data.items()
        except AttributeError:
            # early return because we cannot go further anyway
            return Result.fail(
                Mismatch([], data, f"{repr(data)} does not implement the items method.")
            )

        got = set()
        res = {}
        errors = []

        for key, value in pairs:
            if key not in rec.fields:
                other = closest(rec.fields.keys(), key)
                errors.append(
                    Mismatch([], data, f"{key} is not a valid member of this record. Did you mean {other}")
                )
                continue

            m = self.visit(rec.fields[key], value)
            if m.valid:
                got.add(key)
                res[key] = m.value
            else:
                errors.append(m.error.from_(key))

        for key in set(rec.fields.keys()).difference(got):
            if not rec.fields[key].optional:
                # collect missing members
                errors.append(Mismatch([], data, f"{key} is missing."))
            else:
                default = rec.fields[key].default
                r = self.visit(rec.fields[key], default)
                if not r.valid:
                    return Result.fail(
                        Mismatch([key], default, "Default value is invalid data.")
                    )
                res[key] = r.value

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(res)

    def visit_tuple(self, tup, data) -> Result:
        try:
            elements = list(data)
        except TypeError:
            # early return because we cannot go further anyway
            return Result.fail(Mismatch([], data, f"{repr(data)} is not iterable."))

        res = []
        errors = []

        if len(elements) != len(tup.fields):
            errors.append(
                Mismatch(
                    [],
                    data,
                    f"Expected {len(tup.fields)} elements, but got {len(elements)}.",
                )
            )

        for i, (schema, value) in enumerate(zip(tup.fields, elements)):
            m = self.visit(schema, value)
            if m.valid:
                res.append(m.value)
            else:
                errors.append(m.error.from_(str(i)))

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(res)

    def visit_collection(self, collection, data) -> Result:
        if isinstance(data, str):
            return Result.fail(Mismatch([], data, f"{repr(data)} is a string, not a collection"))

        try:
            col = list(data)
        except AttributeError:
            return Result.fail(Mismatch([], data, f"{repr(data)} is not iterable."))

        validate = []
        errors = []

        for i, e in enumerate(col):
            m = self.visit(collection.element, e)
            if m.valid:
                validate.append(m.value)
            else:
                # collect mismatches in elements
                errors.append(m.error.from_(str(i)))

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(validate)

    def visit_map(self, map_, data) -> Result:
        try:
            pairs = data.items()
        except AttributeError:
            return Result.fail(
                Mismatch([], data, f"{repr(data)} does not implement the items method.")
            )

        res = {}
        errors = []

        for key, value in pairs:
            m = self.visit(map_.element, value)
            if m.valid:
                res[key] = m.value
            else:
                # collect mismatches in elements
                errors.append(m.error.from_(key))

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(res)


@dataclass
class Mismatch:
    "Infos on a mismatch between the schema and a piece of data."

    where: List[str]
    what: object
    why: str

    def from_(self, where):
        return Mismatch([where, *self.where], self.what, self.why)


class Mismatches:
    "A collection of mismatches between schema and data."

    def __init__(self, errors):
        self.errors = []

        for err in errors:
            if isinstance(err, Mismatches):
                self.errors.extend(err.errors)
            else:
                self.errors.append(err)

    def from_(self, where):
        return Mismatches([e.from_(where) for e in self.errors])


class InvalidDataError(ValueError):
    "Exception signaling failure to validate data."

    def __init__(self, mismatch):
        super().__init__(error_message(mismatch))


def error_message(mismatch, no_loc=False):
    "Format the error message for :class:`InvalidDataError`."
    if isinstance(mismatch, Mismatches):
        msg = []
        for mism in mismatch.errors:
            msg.append(error_message(mism, no_loc=no_loc))

        return "\n".join(msg)
    elif no_loc:
        return mismatch.why
    else:
        if None:
            location = ".".join(mismatch.where)
        else:
            location = "[toplevel]"

        return f"In {location}: {mismatch.why}"



"The core definition of a schema."

from dataclasses import dataclass
from typing import Optional, List, Dict, Any
from enum import Enum
from abc import abstractmethod


class Structure:
    """Abstract structure of a node"""

    @abstractmethod
    def accept(self, visitor, *args, **kwargs):
        pass

    @classmethod
    def structure_default(cls):
        raise ValueError(
            f"In {cls} it is an error to set the node optional without a default value."
        )


@dataclass
class Node:
    """Base class for a node of the schema"""

    structure: Structure
    description: Optional[str] = None
    optional: bool = False
    default: Any = None

    def __init__(
        self,
        structure: Structure,
        description: Optional[str] = None,
        optional: bool = False,
        default: Any = None,
    ):
        r"""Create a :class:`Node` instance.

        :param structure: a :class:`Structure` instance describing the node.
        :param describing: (optional, :code:`None`) a human readable description
            as a :code:`str`.
        :param optional: (optional, :code:`False`) wether the node can be
            omited or not when part of a record.
        :param default: (optiona, :code:`None`) a default value used when the
            node is :code:`optional` and omited.
        """
        self.structure = structure
        self.description = description
        self.optional = optional
        self.default = default

        if self.optional and self.default is None:
            self.default = self.structure.structure_default()

    @classmethod
    def of_atom(cls, atom, options=None, **kwargs):
        r"""Constructor for an :class:`Atom` node.

        :param atom: the :class:`AtomType` for the node.
        :param options: options for :class:`Atom` constructor.
        :param \*\*kwargs: kw parameters for :class:`Node.__init__`.
        """
        return cls(Atom(atom, options=options), **kwargs)

    @classmethod
    def of_union(cls, options, **kwargs):
        r"""Constructor for a :class:`Union` node.

        :param options: list of :class:`Node` instances for the alternatives.
        :param \*\*kwargs: kw parameters for :class:`Node.__init__`.
        """
        return cls(Union(options), **kwargs)

    @classmethod
    def of_collection(cls, element, **kwargs):
        r"""Constructor for a :class:`Collection` node.

        :param element: the :class:`Node` instance describing an element of the
            collection.
        :param \*\*kwargs: kw parameters for :class:`Node.__init__`.
        """
        return cls(Collection(element), **kwargs)

    @classmethod
    def of_record(cls, fields, **kwargs):
        r"""Constructor for a :class:`Record` node.

        :param fields: a :code:`dict` of :code:`str`
        :param options: options for :class:`Atom` constructor.
        :param \*\*kwargs: kw parameters for :class:`Node.__init__`.
        """
        return cls(Record(fields), **kwargs)

    @classmethod
    def of_map(cls, element, **kwargs):
        r"""Constructor for a :class:`Map` node.

        :param element: the :class:`Node` instance describing an element of
            the map.
        :param \*\*kwargs: kw parameters for :class:`Node.__init__`.
        """
        return cls(Map(element), **kwargs)

    @classmethod
    def of_tuple(cls, elements, **kwargs):
        r"""Constructor for a :class:`Tuple` node.

        :param elements: list of :class:`Node` describing the elements of the
            tuple.
        :param \*\*kwargs: kw parameters for :class:`Node.__init__`.
        """
        return cls(Tuple(elements), **kwargs)


class AtomType(Enum):
    "Kind of atomic value."

    INT = 0
    FLOAT = 1
    STR = 2
    BOOL = 3
    OPTION = 4

    def name(self, options=None):
        "Return the type name."
        if self == self.INT:
            return "int"
        elif self == self.FLOAT:
            return "float"
        elif self == self.BOOL:
            return "bool"
        elif self == self.STR:
            return "str"
        elif self == self.OPTION:
            return "option(" + ", ".join(repr(s) for s in options) + ")"
        else:
            raise NotImplementedError("This type is not yet named...")


@dataclass
class Atom(Structure):
    """Structure of an atomic piece of data"""

    type_: AtomType
    options: Optional[List[str]] = None

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_atom(self, *args, **kwargs)


@dataclass
class Union(Structure):
    """Structure for the union of several nodes"""

    options: List[Node]

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_union(self, *args, **kwargs)


@dataclass
class Collection(Structure):
    """An ordered collection of similar nodes"""

    element: Node

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_collection(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return []


@dataclass
class Record(Structure):
    """A key-value pair collection.

    Keys are from a defined set and values are defined per-key.
    Keys are supposed to be valid identifiers.
    """

    fields: Dict[str, Node]

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_record(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return {}


@dataclass
class Map(Structure):
    """A key-value pair collection.

    Keys are not restricted and values are of a single type.
    """

    element: Node

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_map(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return {}


@dataclass
class Tuple(Structure):
    """A fixed set of ordered fields.

    Each field can have a different structure.
    """

    fields: List[Node]

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_tuple(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return []



"Miscellaneous utils."


def closest(alternatives, key):
    """Search among :code:`alternatives` the element to :code:`key`.

    Uses :func:`lev_dist` to compute the distance.

    :param alternatives: a :code:`list[str]` of elements to analyse.
    :param key: a :code:`str` key to search for.
    """
    return min(alternatives, key=lambda s: lev_dist(key, s))


def lev_dist(a, b):
    "Compute the Levenstein distance between two :code:`str`."

    len_a = len(a) + 1
    len_b = len(b) + 1

    if len_b > len_a:
        return lev_dist(b, a)

    prev = list(range(len_a))
    current = [0] * len_a

    for j in range(1, len_b):
        current[0] = j
        for i in range(1, len_a):
            if a[i - 1] == b[j - 1]:
                current[i] = prev[i - 1]  # mat[i, j] = mat[i - 1, j - 1]
            else:
                current[i] = min(  # mat[i, j] =
                    current[i - 1] + 1,  # deletion mat[i - 1, j] + 1
                    prev[i] + 1,  # insertion mat[i, j - 1]
                    prev[i - 1] + 1,  # substitution  mat[i - 1, j - 1]
                )

        # swap rows
        current, prev = prev, current

    return prev[-1]


class Result:
    "The Result monad."

    def __init__(self, valid: bool, val):
        if valid:
            self._value = val
            self._error = None
        else:
            self._value = None
            self._error = val

        self.valid = valid

    @classmethod
    def ok(cls, value):
        return cls(True, value)

    @classmethod
    def fail(cls, error):
        return cls(False, error)

    @property
    def value(self):
        assert self.valid, "Try to unwrap invalid result."
        return self._value

    @property
    def error(self):
        assert not self.valid, "Try to unwrap the error of a valid result."
        return self._error

    def __bool__(self):
        return self.valid



"A schema definition module for validation and documentation of structured data."

from .validate import validate
from .describe import describe, search
from .template import template
from .schema import Node, AtomType


__version__ = "0.3.2"
